﻿using System.Collections.Generic;
using System.Linq;

// Filename : EventBase.cs
// Date     : 2024/3/14
//
// Copyright (C) 2024 Grove Design Service All rights reserved.

public class EventBase
{
    public const int FMT_EVT20 = 0;
    public const int FMT_EVT21 = 1;
    public const int FMT_EVT30 = 2;

    public const int IMG_WIDTH  = 640;
    public const int IMG_HEIGHT = 480;

    public const byte ATTR_POSITIVE = 1;
    public const byte ATTR_NEGATIVE = 0;
    public const byte ATTR_SELECTED = 2;
    public const byte ATTR_NONE     = 0xff;

    public const int POL_BOTH     = 0;
    public const int POL_POS_ONLY = 1;
    public const int POL_NEG_ONLY = 2;

    public class Event
    {
        public ushort x;
        public ushort y;
        public byte p;
        public ulong t;

        public Event()
        {
            x = 0;
            y = 0;
            p = 0;
            t = 0;
        }

        public void copy(Event e)
        {
            this.x = e.x;
            this.y = e.y;
            this.p = e.p;
            this.t = e.t;
        }
    }

    // EVT2.0

    private const uint EVT20_MASK      = 0xf0000000;
    private const uint EVT20_OFF       = 0x00000000;
    private const uint EVT20_ON        = 0x10000000;
    private const uint EVT20_TIME_HIGH = 0x80000000;
    private const uint EXT20_TRIGGER   = 0xa0000000;
    private const uint EXT20_OTHERS    = 0xe0000000;
    private const uint EXT20_CONTINUED = 0xf0000000;

    private const uint EVT20_TIME_MASK = 0x0fc00000;
    private const int  EVT20_TIME_SFT  = 22;
    private const uint EVT20_X_MASK    = 0x003ff800;
    private const int  EVT20_X_SFT     = 11;
    private const uint EVT20_Y_MASK    = 0x000007ff;
    private const int  EVT20_Y_SFT     = 0;

    private const uint EVT20_HTIM_MASK = 0x0fffffff;
    private const int  EVT20_HTIM_SFT  = 0;


    // EVT2.1

    private const ulong EVT21_MASK      = 0xf000000000000000;
    private const ulong EVT21_NEG       = 0x0000000000000000;
    private const ulong EVT21_POS       = 0x1000000000000000;
    private const ulong EVT21_TIME_HIGH = 0x8000000000000000;
    private const ulong EXT21_TRIGGER   = 0xa000000000000000;

    private const ulong EVT21_TIME_MASK = 0x0fc0000000000000;
    private const int   EVT21_TIME_SFT  = 54;
    private const ulong EVT21_X_MASK    = 0x003ff80000000000;
    private const int   EVT21_X_SFT     = 43;
    private const ulong EVT21_Y_MASK    = 0x000007ff00000000;
    private const int   EVT21_Y_SFT     = 32;
    private const ulong EVT21_V_MASK    = 0x00000000ffffffff;
    private const int   EVT21_V_SFT     = 0;

    private const ulong EVT21_HTIM_MASK = 0x0fffffff00000000;
    private const int   EVT21_HTIM_SFT  = 32;

    // EVT3.0

    private const ushort EVT30_MASK    = 0xf000;
    private const ushort EVT_ADDR_Y    = 0x0000;
    private const ushort EVT_ADDR_X    = 0x2000;
    private const ushort VECT_BASE_X   = 0x3000;
    private const ushort VECT_12       = 0x4000;
    private const ushort VECT_8        = 0x5000;
    private const ushort EVT_TIME_LOW  = 0x6000;
    private const ushort CONTINUED_4   = 0x7000;
    private const ushort EVT_TIME_HIGH = 0x8000;
    private const ushort EXT_TRIGGER   = 0xa000;
    private const ushort OTHERS        = 0xe000;
    private const ushort CONTINUED_12  = 0xf000;

    public int format = FMT_EVT30;

    // マイクロ秒単位の処理済み時間
    //  （32bitで4294秒=71分までカウントできる。64bitだと580180年）
    public ulong currTime;
    public ulong beginTime; // 最初のイベントの時刻(us)
    public ulong lastTime;  // 最後のイベントの時刻(us)

    // コンストラクタ
    public EventBase(int format)
    {
        // イメージ用のバッファを確保してクリア

        image = new byte[IMG_HEIGHT, IMG_WIDTH];

        clearImage();

        // イベントのリスト生成

        eventList = new List<Event>();

        purgeEvent();

        // デコーダの状態を初期化

        resetDecoder(format);

        beginTime = 0;
        lastTime = 0;

        clearSelectedArea();
    }

    // ===================================
    // イメージに描画

    public byte[,] image;

    // イメージの内容をクリア（イベントがない状態に）する

    public void clearImage()
    {
        for (int y = 0; y < IMG_HEIGHT; y++)
        {
            for (int x = 0; x < IMG_WIDTH; x++)
            {
                image[y, x] = ATTR_NONE;
            }
        }
    }

    // 描画済みのイメージをコピーする
    public void copyImage(byte[,] dest)
    {
        for (int y = 0; y < IMG_HEIGHT; y++)
        {
            for (int x = 0; x < IMG_WIDTH; x++)
            {
                dest[y, x] = image[y, x];
            }
        }
    }

    // 描画済みのイメージをコピーして消す
    public void moveImage(byte[,] dest)
    {
        for (int y = 0; y < IMG_HEIGHT; y++)
        {
            for (int x = 0; x < IMG_WIDTH; x++)
            {
                dest[y, x] = image[y, x];
                image[y, x] = ATTR_NONE;
            }
        }
    }

    // ===================================
    // 保持しているすべてのイベントリストをクリアする

    public List<Event> eventList;
    public int purgeEvent()
    {
        int n = eventList.Count();

        eventList.Clear();

        return n;
    }

    public void addEvent(Event e, int polarity)
    {
        if ((polarity == POL_POS_ONLY) && (e.p == ATTR_NEGATIVE))
        {
            return;
        }

        if ((polarity == POL_NEG_ONLY) && (e.p == ATTR_POSITIVE))
        {
            return;
        }

        Event e0 = new Event();
        e0.p = e.p;
        e0.x = e.x;
        e0.y = e.y;
        e0.t = e.t;
        eventList.Add(e0);
    }

    // ===================================
    // 指定の時間までのRAWデータをデコード

    ulong prevData;      // RAWデータの第1バイト
    int byteCnt;         // デコード済みバイト数
    ushort currCmd = 0;
    ushort prevCmd = 0;
    int currIndex = 0;

    Event e;
    Event e_based;

    // データをリストに追加する

    public List<byte[]> rawDataList = new List<byte[]>();

    public void addRawData(byte[] rawData)
    {
        rawDataList.Add(rawData);
    }

    public bool isDataAvail()
    {
        if (rawDataList.Count == 0)
        {
            return false;
        }

        return true;
    }

    // デコーダの状態を初期値に戻す
    public void resetDecoder(int format)
    {
        this.format = format;

        prevData  = 0;
        byteCnt   = 0;
        currTime  = 0;
        currIndex = 0;
        currCmd   = 0;
        prevCmd   = 0;

        e = new Event();
        e_based = new Event();

        rawDataList.Clear();

        purgeEvent();
    }

    public bool bSelected = false;
    public ushort min_x;
    public ushort min_y;
    public ushort max_x;
    public ushort max_y;

    public void setSelectedArea(ushort minX, ushort minY, ushort maxX, ushort maxY)
    {
        min_x = minX;
        min_y = minY;
        max_x = maxX;
        max_y = maxY;
        bSelected = true;
    }

    public void clearSelectedArea()
    {
        bSelected = false;
    }

    private void updateImage(Event e, int polarity)
    {
        if ((polarity == POL_POS_ONLY) && (e.p == ATTR_NEGATIVE))
        {
            return;
        }

        if ((polarity == POL_NEG_ONLY) && (e.p == ATTR_POSITIVE))
        {
            return;
        }

        if ((e.x < IMG_WIDTH) && (e.y < IMG_HEIGHT))
        {
            if (bSelected &&
                   (e.x >= min_x) &&
                   (e.x <= max_x) &&
                   (e.y >= min_y) &&
                   (e.y <= max_y))
            {
                image[e.y, e.x] = (byte)(e.p | ATTR_SELECTED);
            }
            else
            {
                image[e.y, e.x] = e.p;
            }
        }
    }

    private void updateTime_34(ulong t)
    {
        long currTime_low = (long)(currTime & 0x3ffffffff);
        long currTime_high = (long)(currTime >> 34);

        if ((long)t > currTime_low)
        {
            currTime = ((ulong)currTime_high << 34) | (ulong)t;
        }
        else if ((long)t < (currTime_low - 0x8000))
        {
            currTime = ((ulong)(currTime_high + 1) << 34) | (ulong)t;
        }
    }

    private void updateTime_24(uint t)
    {
        int currTime_low = (int)(currTime & 0x00ffffff);
        int currTime_high = (int)(currTime >> 24);

        if (t > currTime_low)
        {
            currTime = ((ulong)currTime_high << 24) | (ulong)t;
        }
        else if ((int)t < (currTime_low - 0x8000))
        {
            currTime = ((ulong)(currTime_high + 1) << 24) | (ulong)t;
        }
    }

    public bool decodeSub_20(uint d, ulong timeLimit, int polarity)
    {
        switch (d & EVT20_MASK)
        {
            case EVT20_OFF:
                e.t = (e.t & 0x3ffffffc0) | ((d & EVT20_TIME_MASK) >> EVT20_TIME_SFT);
                updateTime_34(e.t);
                e.x = (ushort)((d & EVT20_X_MASK) >> EVT20_X_SFT);
                e.y = (ushort)((d & EVT20_Y_MASK) >> EVT20_Y_SFT);
                e.p = ATTR_NEGATIVE;
                addEvent(e, polarity);      // イベントリストへの追加
                updateImage(e, polarity);   // イメージの更新
                break;
            case EVT20_ON:
                e.t = (e.t & 0x3ffffffc0) | ((d & EVT20_TIME_MASK) >> EVT20_TIME_SFT);
                updateTime_34(e.t);
                e.x = (ushort)((d & EVT20_X_MASK) >> EVT20_X_SFT);
                e.y = (ushort)((d & EVT20_Y_MASK) >> EVT20_Y_SFT);
                e.p = ATTR_POSITIVE;
                addEvent(e, polarity);      // イベントリストへの追加
                updateImage(e, polarity);   // イメージの更新
                break;
            case EVT20_TIME_HIGH:
                e.t = (ulong)((d & EVT20_HTIM_MASK) >> EVT20_HTIM_SFT) << 6;
                updateTime_34(e.t);
                break;
            case EXT20_TRIGGER:
            case EXT20_OTHERS:
            case EXT20_CONTINUED:
                break;
        }

        if (currTime >= timeLimit)
        {
            return true;
        }

        return false;
    }

    public bool decodeUntilTime_20(byte[] rawData, ulong timeTo, int polarity)
    {
        if (rawData == null)
        {
            return false;
        }

        for (int idx = currIndex; idx < rawData.Length; idx++)
        {
            prevData |= (ulong)rawData[idx] << (byteCnt * 8);
            byteCnt++;

            if (byteCnt == 4)
            {
                uint d = (uint)prevData;

                prevData = 0;
                byteCnt = 0;

                if (decodeSub_20(d, timeTo, polarity))
                {
                    currIndex = idx + 1;
                    return true;
                }

            }
        }

        currIndex = 0;

        return false;
    }

    public bool decodeSub_21(ulong d, ulong timeLimit, int polarity)
    {
        switch (d & EVT21_MASK)
        {
            case EVT21_NEG:
                e.t = (e.t & 0x3ffffffc0) | ((d & EVT21_TIME_MASK) >> EVT21_TIME_SFT);
                updateTime_34(e.t);
                e.x = (ushort)((d & EVT21_X_MASK) >> EVT21_X_SFT);
                e.y = (ushort)((d & EVT21_Y_MASK) >> EVT21_Y_SFT);
                e.p = ATTR_NEGATIVE;
                e_based.copy(e);

                for (int i = 0; i < 32; i++)
                {
                    if ((d & (1UL << i)) != 0)
                    {
                        addEvent(e_based, polarity);      // イベントリストへの追加
                        updateImage(e_based, polarity);   // イメージの更新
                    }
                    e_based.x = (ushort)(e_based.x + 1);
                }
                break;
            case EVT21_POS:
                e.t = (e.t & 0x3ffffffc0) | ((d & EVT21_TIME_MASK) >> EVT21_TIME_SFT);
                updateTime_34(e.t);
                e.x = (ushort)((d & EVT21_X_MASK) >> EVT21_X_SFT);
                e.y = (ushort)((d & EVT21_Y_MASK) >> EVT21_Y_SFT);
                e.p = ATTR_POSITIVE;
                e_based.copy(e);

                for (int i = 0; i < 32; i++)
                {
                    if ((d & (1UL << i)) != 0)
                    {
                        addEvent(e_based, polarity);      // イベントリストへの追加
                        updateImage(e_based, polarity);   // イメージの更新
                    }
                    e_based.x = (ushort)(e.x + 1);
                }
                break;
            case EVT21_TIME_HIGH:
                e.t = (uint)((d & EVT21_HTIM_MASK) >> EVT21_HTIM_SFT) << 6;
                updateTime_34(e.t);
                break;
            case EXT21_TRIGGER:
                break;
        }

        if (currTime >= timeLimit)
        {
            return true;
        }

        return false;
    }

    public bool decodeUntilTime_21(byte[] rawData, ulong timeTo, int polarity)
    {
        if (rawData == null)
        {
            return false;
        }

        byteCnt = 0;

        for (int idx = currIndex; idx < rawData.Length; idx++)
        {
            prevData |= (ulong)rawData[idx] << (byteCnt * 8);
            byteCnt++;

            if (byteCnt == 8)
            {
                ulong d = (ushort)prevData;

                prevData = 0;
                byteCnt = 0;

                if (decodeSub_21(d, timeTo, polarity))
                {
                    currIndex = idx + 1;
                    return true;
                }
            }
        }

        currIndex = 0;

        return false;
    }

    public bool decodeSub_30(ushort d, ulong timeLimit, int polarity)
    {
        uint t_new;

        currCmd = (ushort)(d & EVT30_MASK);

        switch (currCmd)
        {
            case EVT_ADDR_Y:
                e.y = (ushort)(d & 0x07ff);
                prevCmd = currCmd;
                break;

            case EVT_ADDR_X:
                e.p = ((d & 0x0800) != 0) ? ATTR_POSITIVE : ATTR_NEGATIVE;
                e.x = (ushort)(d & 0x07ff);

                addEvent(e, polarity);    // イベントリストへの追加
                updateImage(e, polarity); // イメージの更新
                prevCmd = currCmd;
                break;

            case VECT_BASE_X:
                e.p = ((d & 0x0800) != 0) ? ATTR_POSITIVE : ATTR_NEGATIVE;
                e.x = (ushort)(d & 0x07ff);

                // ここではイベントは発生しない、VECT_12またはVECT_8で発生する
                e_based = e;
                prevCmd = currCmd;
                break;

            case VECT_12:
                for (int i = 0; i < 12; i++)
                {
                    if ((d & (1 << i)) != 0)
                    {
                        addEvent(e_based, polarity);      // イベントリストへの追加
                        updateImage(e_based, polarity);   // イメージの更新
                    }
                    e_based.x = (ushort)(e_based.x + 1);
                }
                prevCmd = currCmd;
                break;

            case VECT_8:
                for (int i = 0; i < 8; i++)
                {
                    if ((d & (1 << i)) != 0)
                    {
                        addEvent(e_based, polarity);      // イベントリストへの追加
                        updateImage(e_based, polarity);   // イメージの更新
                    }
                    e_based.x = (ushort)(e_based.x + 1);
                }
                prevCmd = currCmd;
                break;

            case EVT_TIME_LOW:
                e.t = (uint)(e.t & 0x0ffff000) | (uint)(d & 0x000fff);
                updateTime_24((uint)e.t);

                if (currTime >= timeLimit)
                {
                    return true;
                }
                prevCmd = currCmd;

                break;
            case EVT_TIME_HIGH:
                t_new = (uint)(d & 0x000fff);

                e.t = (t_new << 12);
                updateTime_24((uint)e.t);

                if (currTime >= timeLimit)
                {
                    return true;
                }
                prevCmd = currCmd;
                break;

            case CONTINUED_4:
                if (decodeSub_30((ushort)(prevCmd | (d & 0x000f)), timeLimit, polarity))
                {
                    return true;
                }
                break;

            case CONTINUED_12:
                if (decodeSub_30((ushort)(prevCmd | (d & 0x0fff)), timeLimit, polarity))
                {
                    return true;
                }
                break;

            case EXT_TRIGGER:
                prevCmd = currCmd;
                break;

            case OTHERS:
                prevCmd = currCmd;
                break;
        }

        return false;
    }

    public bool decodeUntilTime_30(byte[] rawData, ulong timeTo, int polarity)
    {
        if (rawData == null)
        {
            return false;
        }

        // インデックステーブルからだいたいの位置を取得する(100us単位）

        for (int idx = currIndex; idx < rawData.Length; idx++)
        {
            prevData |= (ulong)rawData[idx] << (byteCnt * 8);
            byteCnt++;

            if (byteCnt == 2)
            {
                ushort d = (ushort)prevData;

                prevData = 0;
                byteCnt = 0;

                if (decodeSub_30(d, timeTo, polarity))
                {
                    currIndex = idx + 1;
                    return true;
                }
            }
        }

        currIndex = 0;

        return false;
    }

    public bool decodeUntilTime(ulong timeTo, int polarity)
    {
        if (rawDataList.Count == 0)
        {
            return false;
        }

        if (currTime > timeTo)
        {
            return true;
        }

        byte[] rawData = rawDataList[0];

        bool bResult = false;

        switch (format)
        {
            case FMT_EVT20: bResult = decodeUntilTime_20(rawData, timeTo, polarity); break;
            case FMT_EVT21: bResult = decodeUntilTime_21(rawData, timeTo, polarity); break;
            case FMT_EVT30: bResult = decodeUntilTime_30(rawData, timeTo, polarity); break;
        }

        if(!bResult)
        {
            rawDataList.RemoveAt(0);
            currIndex = 0;
        }

        return bResult;
    }
}