﻿using System;
using System.IO;                      // FileStream
using System.Threading;               // Thread
using System.Windows;
using System.Windows.Forms;           // FolderBrowserDialog
using System.Windows.Threading;

// Filename : MainWindow.xaml.cs
// Date     : 2024/3/14
//
// Copyright (C) 2024 Grove Design Service All rights reserved.

namespace EBPlayer
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        IniFile ini = new IniFile("EBPlayer");

        private void load_inifile()
        {
            // インストールフォルダ（C:\Program Files\以下）だと、書込み禁止になっていてiniファイルが書き込めない

            String s;

            s = ini.getValueString("view", "frame_rate");
            if (s.Length > 0)
            {
                for (int i = 0; i < cb_frameRate.Items.Count; i++)
                {
                    if (s.Equals(cb_frameRate.Items[i].ToString()))
                    {
                        cb_frameRate.SelectedIndex = i;
                        break;
                    }
                }
            }
        }

        private void save_inifile()
        {
            ini.setValueString("view", "frame_rate", cb_frameRate.SelectedItem.ToString());
        }

        String makerStr;
        String productStr;
        String serialNumberStr;

        // スレッド制御用

        private Thread th_raw;
        private Thread th_view;

        private bool bPlaying = false;
        private bool bStopping = false;
        private bool bClosing = false;

        public MainWindow()
        {
            InitializeComponent();

            // イメージの初期化

            InitializeView();

            // .iniファイルから各種設定の読み取り

            load_inifile();

            // RAWファイル再生スレッドの開始

            th_raw = new Thread(new ThreadStart(thread_raw));
            th_raw.Priority = ThreadPriority.Normal;
            th_raw.Start();

            // ビュー表示スレッドの開始

            th_view = new Thread(new ThreadStart(thread_view));
            th_view.Priority = ThreadPriority.BelowNormal;
            th_view.Start();

            // スレッドが起動するまで待つ
            while (!th_raw.IsAlive) ;
            while (!th_view.IsAlive) ;
        }

        private void MainWindow_Closed(object sender, EventArgs e)
        {
            bClosing = true;

            // 各種設定の.iniファイルへの保存

            save_inifile();
        }

        // ログ
        private void addLog(String s)
        {
            // 画面更新 (スレッドから画面の更新は直接操作できない)
            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
            {
                tb_log.Text += s + "\n";
                tb_log.ScrollToEnd();
            }));
        }

        // ===============================================================
        // ボタンの処理

        private void btn_clearLog_Click(object sender, RoutedEventArgs e)
        {
            tb_log.Text = "";
        }

        private void btn_open_Click(object sender, RoutedEventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.FileName = "";
            ofd.Title = "RAWファイルを指定してください";
            ofd.Filter = "RAW (*.raw)|*.raw";
            ofd.RestoreDirectory = true;
            ofd.CheckFileExists = true;

            if (ofd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    rawFs = new FileStream(ofd.FileName, FileMode.Open);
                }
                catch (Exception ex)
                {
                    return;
                }
            }

            // ヘッダの読み出し

            if (readHeader(rawFs) == false)
            {
                return;
            }

            btn_open.IsEnabled = false;
            btn_stop.IsEnabled = true;
            bPlaying = true;
        }

        private void btn_stop_Click(object sender, RoutedEventArgs e)
        {
            btn_stop.IsEnabled = false;

            bStopping = true;

            Thread.Sleep(100);
        }

        // すべての入力を禁止する

        private void tb_inhibit_all_keys(object sender, System.Windows.Input.KeyEventArgs e)
        {
            e.Handled = true;
        }

        // ===============================================================
        // RAWデータの再生

        private const int RAW_PLAY_BUFFER_SIZE = 128 * 1024;
        ulong targetTime = 0; // 再生開始から処理済みの時間
        DateTime dt_curr;
        EventBase ev = new EventBase(EventBase.FMT_EVT30);

        private String getTimeStr(ulong t)
        {
            String s = "";
            int sec = 0;
            int ms = 0;

            if (t > 1000000)
            {
                sec = (int)(t / 1000000);
                s = sec.ToString() + "s ";
                t = t - (ulong)sec * 1000000;
            }

            if (t > 1000)
            {
                ms = (int)(t / 1000);
                if (sec > 0)
                {
                    s += String.Format("{0:D3}", ms) + "ms ";
                }
                else
                {
                    s += ms.ToString() + "ms ";
                }
                t = t - (ulong)ms * 1000;
            }

            if ((sec > 0) || (ms > 0))
            {
                s += String.Format("{0:D3}", t) + "us";
            }
            else
            {
                s += t.ToString() + "us";
            }

            return s;
        }

        private void thread_raw()
        {
            int length = 0;

            while (!bClosing)
            {
                if (bPlaying)
                {
                    if (bStopping)
                    {
                        // 再生を途中で停止する
                        rawFs.Close();
                        ev.resetDecoder(EventBase.FMT_EVT30);

                        bPlaying = false;
                        bStopping = false;

                        // 画面更新 (スレッドから画面の更新は直接操作できない)
                        this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                        {
                            btn_open.IsEnabled = true;
                            btn_stop.IsEnabled = false;
                        }));
                    }
                    else
                    {
                        DateTime dt_now = DateTime.Now;
                        TimeSpan ts = dt_now - dt_curr;

                        // 実時間が0.1秒経ったら目標時間を更新する

                        if (ts.TotalSeconds >= 0.1)
                        {
                            dt_curr += new TimeSpan(0, 0, 0, 0, 100);
                            targetTime += 100000L;

                            if(targetTime < (ev.currTime - 100000L))
                            {
                                // 画面更新 (スレッドから画面の更新は直接操作できない)
                                this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                                {
                                    tb_status.Text = getTimeStr(targetTime) + "*";
                                }));
                            }
                        }

                        // 時間まで処理していないということはすべてのデータは処理済みなので次を読み出す
                        if (!ev.isDataAvail())
                        {
                            if (rawFs.CanRead)
                            {
                                byte[] rawBuffer = new byte[RAW_PLAY_BUFFER_SIZE];

                                length = rawFs.Read(rawBuffer, 0, rawBuffer.Length);

                                if (length == 0)
                                {
                                    bStopping = true;
                                }
                                else
                                {
                                    if(rawBuffer.Length != length)
                                    {
                                        Array.Resize(ref rawBuffer, length);
                                    }
                                    ev.addRawData(rawBuffer);
                                }
                            }
                            else
                            {
                                bStopping = true;
                            }
                        }

                        if (ev.isDataAvail())
                        {
                            if(ev.decodeUntilTime(targetTime, EventBase.POL_BOTH))
                            {
                                // 目標時刻まで処理済みなら、未処理のデータが残っているので時間待ちする

                                Thread.Sleep(50);
                            }
                        }

                        // ビュー表示スレッドから要求が来た場合には、現時点でのイメージをコピーして、描画中のイメージはクリアする
                        if (bDataRequest && !bDataReady)
                        {
                            ev.moveImage(copiedImage);

                            // 今のところ物体検知はしないので、描画のタイミングでイベントリストもクリアしておく
                            int n = ev.purgeEvent();

                            bDataReady = true;

                            if (ev.currTime < (targetTime + 100000L))
                            {
                                // 画面更新 (スレッドから画面の更新は直接操作できない)
                                this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                                {
                                    tb_status.Text = getTimeStr(ev.currTime);
                                }));
                            }
                        }
                    }
                }
                else
                {
                    // 非再生中はスリープする
                    Thread.Sleep(1000);
                }
            }
        }
    }
}
