﻿using System;
using System.Drawing;                 // Bitmap
using System.Drawing.Imaging;
using System.Threading;               // Thread
using System.Windows;
using System.Windows.Media.Imaging;   // BitmapSource
using System.Windows.Threading;

// Filename : DataView.xaml.cs
// Date     : 2024/3/14
//
// Copyright (C) 2024 Grove Design Service All rights reserved.

namespace EBViewer
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        // イメージ

        private const int IMG_VIEW_WIDTH  = 640;
        private const int IMG_VIEW_HEIGHT = 480;
        private Bitmap bmp_view = new Bitmap(IMG_VIEW_WIDTH, IMG_VIEW_HEIGHT);

        private String [] frameRateStr = { "1 fps", "2 fps", "5 fps", "10 fps", "15 fps" };
        private int [] frameRateSleep = { 1000, 500, 200, 100, 66 };
        private int frameRateIndex = 2;

        private void InitializeView()
        {
            // コンボボックスの初期化

            for (int i = 0; i < frameRateStr.Length; i++)
            {
                cb_frameRate.Items.Add(frameRateStr[i]);
            }
            cb_frameRate.SelectedIndex = 0;

            clearView();
        }
        private void clearView()
        {
            Color cw = Color.FromArgb(255, 255, 255, 255);

            for (int y=0;y< IMG_VIEW_HEIGHT;y++)
            {
                for(int x=0;x<IMG_VIEW_WIDTH;x++)
                {
                    bmp_view.SetPixel(x, y, cw);
                }
            }

            var bitmapData = bmp_view.LockBits(new Rectangle(0, 0, bmp_view.Width, bmp_view.Height), ImageLockMode.ReadOnly, bmp_view.PixelFormat);

            try
            {
                var bitmapSource = BitmapSource.Create(
                    bitmapData.Width,
                    bitmapData.Height,
                    bmp_view.HorizontalResolution,
                    bmp_view.VerticalResolution,
                    System.Windows.Media.PixelFormats.Bgra32,
                    null,
                    bitmapData.Scan0,
                    bitmapData.Stride * bitmapData.Height,
                    bitmapData.Stride);

                img_view.Source = bitmapSource;
            }
            finally
            {
                bmp_view.UnlockBits(bitmapData);
            }
        }

        private void cb_frameRate_SelectionChanged(object sender, System.Windows.Controls.SelectionChangedEventArgs e)
        {
            if(cb_frameRate.SelectedIndex > 0)
            {
                frameRateIndex = cb_frameRate.SelectedIndex;
            }
        }


        // ===============================================================
        // ビュー表示スレッド（データ取得よりもゆっくり行う）

        bool bDataRequest = false;
        bool bDataReady   = false;

        byte[,] copiedImage = new byte[IMG_VIEW_HEIGHT, IMG_VIEW_WIDTH];
        private void thread_view()
        {
            while (!bClosing)
            {
                if(bRunning)
                {
                    if(bDataRequest)
                    {
                        if (bDataReady)
                        {
                            bDataRequest = false;
                            bDataReady   = false;

                            // 画面更新 (スレッドから画面の更新は直接操作できない)
                            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                            {
                                Color cw    = Color.FromArgb(255, 255, 255, 255);
                                Color c_pos = Color.FromArgb(255, 30, 37, 52);
                                Color c_neg = Color.FromArgb(255, 64, 126, 200);

                                for (int y = 0; y < IMG_VIEW_HEIGHT; y++)
                                {
                                    for (int x = 0; x < IMG_VIEW_WIDTH; x++)
                                    {
                                        Rect r = new Rect(x, y, 1.0, 1.0);

                                        byte attr = copiedImage[y, x];
                                        if (attr == EventBase.ATTR_POSITIVE)
                                        {
                                            bmp_view.SetPixel(x, y, c_pos);
                                        }
                                        else if (attr == EventBase.ATTR_NEGATIVE)
                                        {
                                            bmp_view.SetPixel(x, y, c_neg);
                                        }
                                        else
                                        {
                                            bmp_view.SetPixel(x, y, cw);
                                        }
                                    }
                                }

                                var bitmapData = bmp_view.LockBits(new Rectangle(0, 0, bmp_view.Width, bmp_view.Height), ImageLockMode.ReadOnly, bmp_view.PixelFormat);

                                try
                                {
                                    var bitmapSource = BitmapSource.Create(
                                        bitmapData.Width,
                                        bitmapData.Height,
                                        bmp_view.HorizontalResolution,
                                        bmp_view.VerticalResolution,
                                        System.Windows.Media.PixelFormats.Bgra32,
                                        null,
                                        bitmapData.Scan0,
                                        bitmapData.Stride * bitmapData.Height,
                                        bitmapData.Stride);

                                    img_view.Source = bitmapSource;
                                }
                                finally
                                {
                                    bmp_view.UnlockBits(bitmapData);
                                }
                            }));

                            // フレームレートの設定によって、待機時間を変える
                            Thread.Sleep(frameRateSleep[frameRateIndex]);
                        }
                        else
                        {
                            // データが書き込まれるまで待つ
                            Thread.Sleep(1);
                        }
                    }
                    else
                    {
                        bDataRequest = true;
                        Thread.Sleep(1);
                    }
                } else
                {
                    Thread.Sleep(100);
                }
            }
        }
    }
}
