﻿using System;
using System.Runtime.InteropServices; // DLL Import

// Filename : EVC.cs
// Date     : 2024/2/21
//
// Copyright (C) 2024 Grove Design Service All rights reserved.
//
// 本プログラムはイベントベースカメラにアクセスするためのC# 用クラスです。
// EVC.dll と共に使用します。DLLで使用されいているアンマネージドな引数をマネージドに変換します。
// 本プログラムは改変のない限り、自由に使用、再配布できますが、著作権は放棄されていません。


class EVC
{
    //  ｵｰﾌﾟﾝ/ｸﾛｰｽﾞ関数
    [DllImport("EVC.dll", EntryPoint = "EVC_UsbOpen")]
    public static extern bool EVC_UsbOpen(out UInt64 pdFoundHanlde, out UInt64 pdWinUSBHandle);

    [DllImport("EVC.dll", EntryPoint = "EVC_UsbClose")]
    public static extern bool EVC_UsbClose(UInt64 dFoundHanlde, UInt64 dWinUSBHandle);

    [DllImport("EVC.dll", EntryPoint = "EVC_GetDeviceInfo")]
    public static extern bool EVC_GetDeviceInfo(UInt64 dFoundHanlde, IntPtr pMakerStr, IntPtr pProductStr);

    private const int MAX_MAKER_STR_LEN = 127; // char型は2バイトなので127文字
    private const int MAX_PRODUCT_STR_LEN = 127; // char型は2バイトなので127文字

    [DllImport("EVC.dll", EntryPoint = "EVC_GetSpeed")]
    public static extern bool EVC_GetSpeed(UInt64 dFoundHanlde, IntPtr pDeviceSpeed);

    public const int LOW_SPEED = 1;
    public const int FULL_SPEED = 2;
    public const int HIGH_SPEED = 3;

    [DllImport("EVC.dll", EntryPoint = "EVC_GetSerialNumber")]
    public static extern bool EVC_GetSerialNumber(UInt64 dFoundHanlde, IntPtr pSerialStr);

    [DllImport("EVC.dll", EntryPoint = "EVC_Connect")]
    public static extern bool EVC_Connect(UInt64 dFoundHanlde);

    [DllImport("EVC.dll", EntryPoint = "EVC_Start")]
    public static extern bool EVC_Start(UInt64 dFoundHanlde);

    [DllImport("EVC.dll", EntryPoint = "EVC_Stop")]
    public static extern bool EVC_Stop(UInt64 dFoundHanlde);

    [DllImport("EVC.dll", EntryPoint = "EVC_ReadData")]
    public static extern bool EVC_ReadData(UInt64 dFoundHanlde, out int length, IntPtr pData);


    // メンバ関数

    private UInt64 dFoundHandle = 0;
    private UInt64 dWinUSBHandle = 0;

    // スレッド間排他制御

    public Object cmd_lockObject;
    public Object dat_lockObject;
    public const int MINIMUM_BUFFER_SIZE = 0x20000;
    IntPtr pDataBuffer;

    public EVC()
    {
        cmd_lockObject = new object();
        dat_lockObject = new object();

        pDataBuffer = Marshal.AllocCoTaskMem(EVC.MINIMUM_BUFFER_SIZE + 4096);
    }

    ~EVC()
    {
        Marshal.FreeCoTaskMem(pDataBuffer);
    }

    public bool UsbOpen()
    {
        // すでにオープンされていたら重複してオープンできない

        if (dWinUSBHandle != 0)
        {
            return false;
        }

        UInt64 pdFoundHandle;
        UInt64 pdWinUSBHandle;

        bool val = EVC_UsbOpen(out pdFoundHandle, out pdWinUSBHandle);

        if (val == true)
        {
            dFoundHandle = pdFoundHandle;
            dWinUSBHandle = pdWinUSBHandle;
        }
        else
        {
            dFoundHandle = (UInt64)0;
            dWinUSBHandle = (UInt64)0;
        }

        return val;
    }

    public void UsbClose()
    {
        EVC_UsbClose(dFoundHandle, dWinUSBHandle);

        dFoundHandle = 0;
        dWinUSBHandle = 0;
    }

    private string IntPtr_to_String(IntPtr s)
    {
        // Intptr型のポインタをbyte[]配列にコピー

        char[] buf = new char[MAX_MAKER_STR_LEN];
        int length = 0;

        Marshal.Copy(s, buf, 0, buf.Length);

        for (int i=0;i<buf.Length;i++)
        {
            if(buf[i] == '\0')
            {
                length = i;
                break;
            }
        }

        // char[] 配列をstring型に変換

//    　string str = System.Text.Encoding.ASCII.GetString(buf);
        string str = new string(buf);

        return str.Substring(0,length);
    }

    public bool GetDeviceInfo(out string makerStr, out string productStr)
    {
        bool bSuccess = false;

        lock (cmd_lockObject)
        {
            // 文字読み出し用のバッファを確保

            IntPtr intPtrMaker = Marshal.AllocCoTaskMem(MAX_MAKER_STR_LEN * sizeof(char));
            IntPtr intPtrProduct = Marshal.AllocCoTaskMem(MAX_PRODUCT_STR_LEN * sizeof(char));

            bSuccess = EVC_GetDeviceInfo(dWinUSBHandle, intPtrMaker, intPtrProduct);

            if (bSuccess)
            {
                makerStr   = IntPtr_to_String(intPtrMaker);
                productStr = IntPtr_to_String(intPtrProduct);
            }
            else
            {
                makerStr   = "";
                productStr = "";
            }

            Marshal.FreeCoTaskMem(intPtrMaker);
            Marshal.FreeCoTaskMem(intPtrProduct);
        }

        return bSuccess;
    }

    public bool GetSpeed(out int speed)
    {
        bool bSuccess = false;

        lock (cmd_lockObject)
        {
            IntPtr pTemp = Marshal.AllocCoTaskMem(sizeof(byte));

            bSuccess = EVC_GetSpeed(dWinUSBHandle, pTemp);

            if (bSuccess)
            {
                // Intptr型のポインタをbyte[]配列にコピー

                byte[] buf = new byte[1];

                Marshal.Copy(pTemp, buf, 0, buf.Length);

                speed = (int)buf[0];
            }
            else
            {
                speed = 0;
            }

            Marshal.FreeCoTaskMem(pTemp);
        }

        return bSuccess;
    }

    public bool GetSerialNumber(out string sSerialNumber)
    {
        bool bSuccess = false;

        lock (cmd_lockObject)
        {
            IntPtr pSerial = Marshal.AllocCoTaskMem(32);

            bSuccess = EVC_GetSerialNumber(dWinUSBHandle, pSerial);

            if (bSuccess)
            {
                // 一旦byte型の配列にコピーする
                byte[] buf = new byte[8];

                Marshal.Copy(pSerial, buf, 0, buf.Length);

                // String型に変換
                sSerialNumber = System.Text.Encoding.ASCII.GetString(buf);
            }
            else
            {
                sSerialNumber = "";
            }

            Marshal.FreeCoTaskMem(pSerial);
        }

        return bSuccess;
    }

    public bool Connect()
    {
        bool bSuccess = false;

        lock (cmd_lockObject)
        {
            bSuccess = EVC_Connect(dWinUSBHandle);
        }

        return bSuccess;
    }

    public bool Start()
    {
        bool bSuccess = false;

        lock (cmd_lockObject)
        {
            bSuccess = EVC_Start(dWinUSBHandle);
        }

        return bSuccess;
    }

    public bool Stop()
    {
        bool bSuccess = true;

        lock (cmd_lockObject)
        {
            bSuccess = EVC_Stop(dWinUSBHandle);
        }

        return bSuccess;
    }

    public bool ReadData(out byte[] pData)
    {
        bool bSuccess = false;
        int len = 0;

        lock (dat_lockObject)
        {
            if (pDataBuffer != IntPtr.Zero)
            {
                bSuccess = EVC_ReadData(dWinUSBHandle, out len, pDataBuffer);

                if (bSuccess)
                {
                    pData = new byte[len];

                    Marshal.Copy(pDataBuffer, pData, 0, pData.Length);
                }
                else
                {
                    pData = new byte[0];
                }
            }
            else
            {
                pData = new byte[0];
            }
        }

        return bSuccess;
    }
}

