﻿using System;
using System.Collections.Generic;
using System.IO;                      // FileStream
using System.Threading;               // Thread
using System.Windows;
using System.Windows.Threading;

// Filename : MainWindow.xaml.cs
// Date     : 2024/3/14
//
// Copyright (C) 2024 Grove Design Service All rights reserved.

namespace EBViewer
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        IniFile ini = new IniFile("EBViewer");

        private void load_inifile()
        {
            // インストールフォルダ（C:\Program Files\以下）だと、書込み禁止になっていてiniファイルが書き込めない

            String s;

            s = ini.getValueString("view", "frame_rate");
            if (s.Length > 0)
            {
                for (int i = 0; i < cb_frameRate.Items.Count; i++)
                {
                    if (s.Equals(cb_frameRate.Items[i].ToString()))
                    {
                        cb_frameRate.SelectedIndex = i;
                        break;
                    }
                }
            }

            s = ini.getValueString("raw", "saveToRaw");
            cb_saveToRaw.IsChecked = s.Equals("yes") ? true : false;

            s = ini.getValueString("raw", "saveFolder");
            tb_saveFolder.Text = s;
        }

        private void save_inifile()
        {
            ini.setValueString("view", "frame_rate", cb_frameRate.SelectedItem.ToString());
            ini.setValueString("raw", "saveToRaw", (cb_saveToRaw.IsChecked == true) ? "yes" : "no");
            ini.setValueString("raw", "saveFolder", tb_saveFolder.Text);
        }

        // イベントベースカメラへのアクセスクラスを生成

        private EVC evc = new EVC();

        String makerStr;
        String productStr;
        String serialNumberStr;

        // スレッド制御用

        private Thread th_data;
        private Thread th_view;

        private bool bOpened = false;
        private bool bRunning = false;
        private bool bStopping = false;
        private bool bClosing = false;

        public MainWindow()
        {
            InitializeComponent();

            // イメージの初期化

            InitializeView();

            // .iniファイルから各種設定の読み取り

            load_inifile();

            // データ取得スレッドの開始

            th_data = new Thread(new ThreadStart(thread_data));
            th_data.Priority = ThreadPriority.Normal;
            th_data.Start();

            // ビュー表示スレッドの開始

            th_view = new Thread(new ThreadStart(thread_view));
            th_view.Priority = ThreadPriority.BelowNormal;
            th_view.Start();

            // スレッドが起動するまで待つ
            while (!th_data.IsAlive) ;
            while (!th_view.IsAlive) ;
        }

        private void MainWindow_Closed(object sender, EventArgs e)
        {
            bClosing = true;

            // 各種設定の.iniファイルへの保存

            save_inifile();
        }

        // ログ
        private void addLog(String s)
        {
            // 画面更新 (スレッドから画面の更新は直接操作できない)
            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
            {
                tb_log.Text += s + "\n";
                tb_log.ScrollToEnd();
            }));
        }

        // ===============================================================
        // ボタンの処理

        private void btn_clearLog_Click(object sender, RoutedEventArgs e)
        {
            tb_log.Text = "";
        }

        private void btn_start_Click(object sender, RoutedEventArgs e)
        {
            // チェックボックスの状態をメンバに反映する（他スレッドから参照するため）

            bSaveToRaw = (cb_saveToRaw.IsChecked == true) ? true : false;

            if (evc.Start() == false)
            {
                addLog("-E- Start failed");
                return;
            }

            addLog("-I- Start");

            ev.resetDecoder(EventBase.FMT_EVT30);

            if (bSaveToRaw)
            {
                createRawFile();
            }

            dt_curr = DateTime.Now;
            targetTime = 0;

            btn_start.IsEnabled = false;
            btn_stop.IsEnabled = true;
            bRunning = true;
        }

        private void btn_stop_Click(object sender, RoutedEventArgs e)
        {
            btn_stop.IsEnabled = false;

            // 先に入力を止めてしまう（注意：カメラ内にデータが少し残る）

            bStopping = true;

            Thread.Sleep(100);

            if (evc.Stop() == false)
            {
                addLog("-E- Stop() failed");
                return;
            }

            addLog("-I- Stop()");

            // 一旦カメラをクローズしないと、次回再生時に時間が0から始まらない
            close_camera();
            bOpened = false;
        }

        // すべての入力を禁止する

        private void tb_inhibit_all_keys(object sender, System.Windows.Input.KeyEventArgs e)
        {
            e.Handled = true;
        }


        // ===============================================================
        // データ読み出しスレッド

        private bool open_camera()
        {
            if (evc.UsbOpen() == false)
            {
                return false;
            }

            addLog("-I- Usb_Open()");


            if (evc.GetDeviceInfo(out makerStr, out productStr) == false)
            {
                addLog("-I- Get_DeviceInfo() failed");
                return false;
            }

            addLog("-I- Get_DeviceInfo()");
            addLog("maker   : " + makerStr);
            addLog("product : " + productStr);

            int speed;

            if (evc.GetSpeed(out speed) == false)
            {
                addLog("-I- Get_Speed() failed");
                return false;
            }

            addLog("-I- Get_Speed()");
            switch (speed)
            {
                case EVC.LOW_SPEED: addLog("Device connected in Low-Speed"); break;
                case EVC.FULL_SPEED: addLog("Device connected in Full-Speed"); break;
                case EVC.HIGH_SPEED: addLog("Device connected in High-Speed"); break;
            }

            if (evc.GetSerialNumber(out serialNumberStr) == false)
            {
                addLog("-I- GetSerialNumber() failed");
                return false;
            }
            addLog("-I- GetSerialNumber()");
            addLog("SerialNumber : " + serialNumberStr);

            if (evc.Connect() == false)
            {
                addLog("-I- Connect() failed");
                return false;
            }
            addLog("-I- Connect()");

            // 画面更新 (スレッドから画面の更新は直接操作できない)
            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
            {
                btn_start.IsEnabled = true;
                btn_stop.IsEnabled = true;
            }));

            return true;
        }

        private void close_camera()
        {
            evc.UsbClose();

            addLog("-I- Close()");

            // 画面更新 (スレッドから画面の更新は直接操作できない)
            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
            {
                btn_start.IsEnabled = false;
                btn_stop.IsEnabled = false;
            }));
        }

        EventBase ev = new EventBase(EventBase.FMT_EVT30);
        DateTime dt_curr = DateTime.Now;
        ulong targetTime = 0;

        private String getTimeStr(ulong t)
        {
            String s = "";
            int sec = 0;
            int ms = 0;

            if (t > 1000000)
            {
                sec = (int)(t / 1000000);
                s = sec.ToString() + "s ";
                t = t - (ulong)sec * 1000000;
            }

            if (t > 1000)
            {
                ms = (int)(t / 1000);
                if (sec > 0)
                {
                    s += String.Format("{0:D3}", ms) + "ms ";
                }
                else
                {
                    s += ms.ToString() + "ms ";
                }
                t = t - (ulong)ms * 1000;
            }

            if ((sec > 0) || (ms > 0))
            {
                s += String.Format("{0:D3}", t) + "us";
            }
            else
            {
                s += t.ToString() + "us";
            }

            return s;
        }

        private void thread_data()
        {
            byte[] pDataBuffer;

            while (!bClosing)
            {
                if (!bOpened)
                {
                    // カメラが接続されていなければ接続を試す(1秒に1回）
                    if (open_camera())
                    {
                        bOpened = true;
                    }
                    else
                    {
                        Thread.Sleep(1000);
                    }
                }
                else
                {
                    if (bRunning)
                    {
                        TimeSpan ts = DateTime.Now - dt_curr;

                        // 実時間が0.1秒経ったら目標時間を更新する

                        if (ts.TotalSeconds >= 0.1)
                        {
                            dt_curr += new TimeSpan(0, 0, 0, 0, 100);
                            targetTime += 100000L;
                        }

                        if (evc.ReadData(out pDataBuffer) == false)
                        {
                            // エラーが発生したら一旦WinUsbをクローズする
                            close_camera();

                            // 画面更新 (スレッドから画面の更新は直接操作できない)
                            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                            {
                                addLog("ReadData() failed");
                            }));
                        }
                        else
                        {
                            if (pDataBuffer.Length > 0) // 読み出すべきデータがあるとき
                            {
                                writeRawFile(pDataBuffer);
                                ev.addRawData(pDataBuffer);
                            }
                            else
                            {
                                Thread.Sleep(100);
                            }
                        }

                        if(ev.isDataAvail())
                        {
                            ev.decodeUntilTime(targetTime, EventBase.POL_BOTH);
                        }

                        // 画面更新 (スレッドから画面の更新は直接操作できない)
                        this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                        {
                            tb_status.Text = getTimeStr(ev.currTime);
                        }));

                        // ビュー表示スレッドから要求が来た場合には、現時点でのイメージをコピーして、描画中のイメージはクリアする
                        if (bDataRequest && !bDataReady)
                        {
                            ev.moveImage(copiedImage);

                            // 今のところ物体検知はしないので、描画のタイミングでイベントリストもクリアしておく
                            int n = ev.purgeEvent();

                            bDataReady = true;

                            // 画面更新 (スレッドから画面の更新は直接操作できない)
                            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                            {
                                tb_status.Text = getTimeStr(ev.currTime);
                            }));
                        }

                        if (bStopping) //　停止指示があれば停止
                        {
                            bRunning = false;
                            bStopping = false;
                            closeRawFile();

                            // 画面更新 (スレッドから画面の更新は直接操作できない)
                            this.Dispatcher.BeginInvoke(DispatcherPriority.Background, new Action(() =>
                            {
                                btn_start.IsEnabled = true;
                            }));
                            addLog("-I- Stopped");
                        }
                    }
                    else
                    {
                        // 非キャプチャー中は定期的にアクセスして取り外しを検知する

                        int speed;

                        if (evc.GetSpeed(out speed) == false)
                        {
                            close_camera();
                            bOpened = false;
                        }
                        else
                        {
                            Thread.Sleep(1000);
                        }
                    }
                }
            }
        }
    }
}
